# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
External clustering evaluation metrics.

This module contains evaluation metrics that require ground truth
labels in addition to the cluster assignments.
These metrics are typically used to compare clustering results
against known labels.
"""

import numpy as np
import pandas as pd
from scipy.stats import chi2_contingency
from sklearn import metrics
from sklearn.metrics import confusion_matrix

from .base import EvaluationMetric
from .utils import _get_cont_table


class AdjustedMutualInformationScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        ## AMI: Function that measures how well the optimal assignment corresponds to the one returned by the algorithm.
        # Based on entropy. Normalized and symmetric.
        # Bad assignments can take negative values, optimal assignment -> 1.
        return metrics.adjusted_mutual_info_score(self.true_labels, assigned_clusters)


class AdjustedRandIndexScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        ## Coefficient that assigns 1. to the optimal prediction and values close to 0 the more random the predictions are
        return metrics.adjusted_rand_score(self.true_labels, assigned_clusters)


class VMeasureScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        ## Combination of two criteria: homogeneity and completeness.
        # Homogeneity: each cluster contains only values corresponding to one label.
        # Completeness: all values corresponding to the same label are assigned to the same cluster.
        # The v-measure is a combination of both criteria, weighted by a beta value (less than one for more homogeneity,
        # greater than one for more completeness).
        # It's normalized and symmetric. Random assignment does not get a score of 0. (!)
        return metrics.v_measure_score(self.true_labels, assigned_clusters)


class FowlkesMallowsScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        ## Geometric mean of precision and recall.
        # Random assignments have a value close to 0.
        # Normalized.
        return metrics.fowlkes_mallows_score(self.true_labels, assigned_clusters)


class FScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        return metrics.f1_score(self.true_labels, assigned_clusters, average="weighted")


class NMIScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        return metrics.normalized_mutual_info_score(self.true_labels, assigned_clusters)


class HScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        true_labels = self.true_labels
        predicted_cluster = assigned_clusters
        confusion = confusion_matrix(true_labels, predicted_cluster)
        # Number of clusters
        k = confusion.shape[0]
        # Number of elements in the dataset
        n = np.sum(confusion)
        # Calculate the criterion H
        ch = 1 - (1 / n) * np.sum(np.max(confusion, axis=1))
        return ch


class Chi2(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        table1 = _get_cont_table(assigned_clusters, self.true_labels, round=False)
        table2 = _get_cont_table(self.true_labels, assigned_clusters, round=False)
        chi2_1 = chi2_contingency(table1)[0]
        chi2_2 = chi2_contingency(table2)[0]
        return np.sum([chi2_1, chi2_2])


class DobPertScore(EvaluationMetric):
    def compute(self, assigned_clusters, unlabeled_data=None):
        cont_clust = _get_cont_table(assigned_clusters, self.true_labels, round=False)
        cont_event = _get_cont_table(self.true_labels, assigned_clusters, round=False)

        heuristic_table = pd.DataFrame(
            columns=["event_code", "heuristic_value"]
        ).set_index("event_code")
        for event in cont_event.index:
            clust_table = pd.DataFrame(
                columns=["cluster", "heuristic_value"]
            ).set_index("cluster")
            for clust in cont_clust.index:
                clust_table.loc[clust] = (
                    cont_clust.loc[clust, event] + cont_event.loc[event, clust]
                )

            if (clust_table["heuristic_value"] > 48).any():
                heuristic_table.loc[event] = clust_table["heuristic_value"].max()
        return heuristic_table["heuristic_value"].sum()



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gaufs import DataGenerator
from gaufs import Gaufs
from gaufs import read_labeled_data_csv
from gaufs.evaluation_metrics import AdjustedMutualInformationScore

def main():
    seed = 0
    # Generate data with corners distribution and save it to a CSV file
    data_with_labels = DataGenerator.generate_data_corners(
        num_useful_features=2,
        num_samples_per_cluster=10,
        num_dummy_unif=1,
        num_dummy_beta=1,
        alpha_param=2,
        beta_param=3,
        output_path="./datasets/corners_3clusters.csv",
        seed=seed,
    )

    # Read the data
    unlabeled_data, true_labels = read_labeled_data_csv(
        "./datasets/corners_3clusters.csv"
    )

    # Instantiate GAUFS
    gaufs = Gaufs(seed=seed)
    # Set the unlabeled data
    gaufs.set_unlabeled_data(unlabeled_data)
    # We recomend to leave ngen and npop with the default values but for quick testing we change them
    gaufs.ngen = 2
    gaufs.npop = 250
    # Run GAUFS
    gaufs.run()

    # Comparison with external metric
    gaufs.get_plot_comparing_solution_with_another_metric(
        AdjustedMutualInformationScore(true_labels=true_labels),
        true_number_of_labels=len(set(true_labels)),
    )


if __name__ == "__main__":
    main()



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
kmeans.py

KMeans clustering experiment for GAUFS.
Implements the KmeansExperiment class, which
extends ClusteringExperiment using KMeans.
"""

from sklearn.cluster import KMeans

from .base import ClusteringExperiment


class KmeansExperiment(ClusteringExperiment):
    """
    K-means clustering experiment.

    Parameters
    ----------
    unlabeled_data : pandas.DataFrame or pandas.Series, optional
        The unlabeled dataset to be clustered. If provided, a copy is stored.
    seed : int, optional
        Random state seed for reproducibility. If None, results are non-deterministic.

    Attributes
    ----------
    n_clusters : int or None
        Number of clusters to form. Must be set before calling run().
    seed : int or None
        Random state seed used for the KMeans algorithm.

    """

    def __init__(self, unlabeled_data=None, seed=None):
        """
        Initialize a KmeansExperiment instance.

        Parameters
        ----------
        unlabeled_data : pandas.DataFrame or pandas.Series, optional
            The unlabeled dataset to be clustered.
        seed : int, optional
            Random state seed for reproducibility.
        """
        if unlabeled_data is not None:
            self.unlabeled_data = unlabeled_data.copy()
        self.n_clusters = None
        self.seed = seed

    def run(self):
        """
        Execute K-means clustering.

        Creates a KMeans instance with the specified number of clusters
        and optional random state, then runs the clustering.

        """
        self.algorithm = (
            KMeans(n_clusters=self.n_clusters, random_state=self.seed)
            if self.seed != None
            else KMeans(n_clusters=self.n_clusters)
        )
        super().run()



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
base.py

Base class for clustering experiments in GAUFS.
Defines the common interface and functionality for all clustering
algorithm implementations.
Users can extend this class to implement custom clustering algorithms.
"""

import copy
import pandas as pd


class ClusteringExperiment:
    """
    Base class for clustering experiments.

    This class defines a common interface and shared functionality for
    unsupervised clustering experiments. Concrete subclasses are expected
    to configure the `algorithm` attribute with a clustering model that
    implements `fit()` and exposes `labels_` after fitting.

    Attributes
    ----------
    unlabeled_data : pandas.DataFrame or array-like
        Input data used for clustering.
    algorithm : object
        Clustering algorithm instance (e.g., KMeans, DBSCAN).
    assigned_clusters : pandas.Series or array-like
        Cluster labels assigned after running the experiment.
    """

    def __init__(self):
        """
        Initialize a new ClusteringExperiment instance.

        Sets all attributes to None, including unlabeled_data, algorithm instance,
        and assigned_clusters results.

        """
        self.unlabeled_data = None
        # Algorithm instance
        self.algorithm = None
        # Result: assigned clusters
        self.assigned_clusters = None

    def set_unlabeled_data(self, unlabeled_data):
        """
        Set the unlabeled data for clustering.

        Parameters
        ----------
        unlabeled_data : pandas.DataFrame or pandas.Series
            The unlabeled dataset to be clustered. A copy is stored to avoid
            modifying the original data.

        """
        self.unlabeled_data = unlabeled_data.copy()

    def run(self):
        """
        Execute the clustering algorithm on the unlabeled data.

        Fits the clustering algorithm to the unlabeled data and stores the
        resulting cluster labels in the assigned_clusters attribute.

        Notes
        -----
        This method assumes that both unlabeled_data and algorithm have been
        properly set. The cluster labels are stored in :attr:`assigned_clusters`.

        See Also
        --------
        assigned_clusters : The attribute where cluster labels are stored
        """
        x_clust = self.unlabeled_data
        self.algorithm.fit(x_clust)
        self.assigned_clusters = self.algorithm.labels_

    def get_dist_samples(self):
        """
        Get the distribution of samples across clusters.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with cluster labels as index and two columns:
            - First column: absolute count of samples in each cluster
            - 'relative': percentage of samples in each cluster
            The DataFrame is sorted by cluster label index.

        Notes
        -----
        This method should be called after run() has been executed to ensure
        assigned_clusters contains valid cluster labels.

        """
        value_counts = pd.DataFrame(self.assigned_clusters.value_counts())
        value_counts["relative"] = (
            self.assigned_clusters.value_counts(normalize=True) * 100
        )
        return value_counts.sort_index()

    def get_data_with_clusters(self):
        """
        Get the unlabeled data with assigned cluster labels.

        Returns
        -------
        pandas.DataFrame
            Copy of unlabeled data with added 'cluster' column.

        Raises
        ------
        ValueError
            If clustering has not been run yet.

        """
        data_with_clusters = self.unlabeled_data.copy()
        if self.assigned_clusters is None:
            raise ValueError(
                "Clustering has not been run yet. Please run the clustering algorithm first."
            )
        data_with_clusters["cluster"] = self.assigned_clusters
        return data_with_clusters

    def __copy__(self):
        """Shallow copy"""
        new_instance = type(self)()
        new_instance.__dict__.update(self.__dict__)
        return new_instance

    def __deepcopy__(self, memo):
        """Deep copy"""
        new_instance = type(self)()
        memo[id(self)] = new_instance
        for key, value in self.__dict__.items():
            setattr(new_instance, key, copy.deepcopy(value, memo))
        return new_instance


# To do: explore other clustering algorithms such as DBSCAN, MeanShift, Spectral Clustering, etc.
